# multiplexer.tcl -- one to many comunication with sockets
#
#	Implementation of a one-to-many multiplexer in Tcl.

# Copyright (c) 2001 by David N. Welton <davidw@dedasys.com>

# This file may be distributed under the same terms as Tcl.

# $Id$

package provide multiplexer 0.1

namespace eval multiplexer {
    # list of connected clients
    array set clients {}

    # filters to run at access (socket accept) time
    set accessfilters {}

    # filters to run on data
    set filters {}

    # hook to run at exit time
    set exitfilters {}

    # config options
    array set config {}
    set config(sendtoorigin) 0
    set config(buffering) line
    set config(debug) 0
}

# AddAccessFilter --

# Command to add an access filter that will be called like so:

# AccessFilter chan clientaddress clientport

# Arguments:

# function: proc to filter access to the multiplexer.  Takes chan,
# clientaddress and clientport arguments.  Returns 0 on success, -1 on
# failure.

proc multiplexer::AddAccessFilter { function } {
    variable accessfilters
    lappend accessfilters $function
}

# AddFilter --

# Command to add a filter for data that passes through the
# multiplexer.  The filter proc is called like this:

# Filter data chan clientaddress clientport

# Arguments:

# function: proc to filter data that arrives to the multiplexer.
# Takes data, chan, clientaddress, and clientport arguments.  Returns
# filtered version of data.

proc multiplexer::AddFilter { function } {
    variable filters
    lappend filters $function
}

# AddExitFilter --

# Adds filter to be run when client socket generates an EOF condition.
# ExitFilter functions look like the following:

# ExitFilter chan clientaddress clientport

# Arguments:

# function: hook to be run when clients exit by generating an EOF.
# Takes chan, clientaddress and clientport arguments, and returns
# nothing.

proc multiplexer::AddExitFilter { function } {
    variable exitfilters
    lappend exitfilters $function
}

# DelClient --

# Deletes a client from the client list, and runs exit filters.

# Arguments:

# chan: channel that is closed.

# client: address of client

# clientport: port number of client.

proc multiplexer::DelClient { chan client clientport } {
    variable clients
    variable exitfilters
    variable config
    foreach ef $exitfilters {
	catch {
	    $ef $chan $client $clientport
	} err
        if { $config(debug) > 0 } {
	    puts stderr $err
	}
    }
    unset clients($chan)
    close $chan
}

# GetData --

# Main function of multiplexer which gets data from clients and
# redistributes it.

# Arguments:

# chan: open channel

# client: client address

# clientport: port number of client

proc multiplexer::GetData { chan client clientport } {
    variable filters
    variable clients
    variable config
    if { ! [ eof $chan ] } {
	gets $chan data
	if { $config(debug) > 1 } {
	    puts stderr "Tcl chan $chan from host $client and port $clientport sends: $data"
	}
	# do data filters
	foreach f $filters {
	    catch {
		set data [ $f $data $chan $client $clientport ]
	    } err
            if { $config(debug) > 1 } {
		puts stderr $err
	    }
	}
	set chans [ array names clients ]
	foreach c [array names clients] {
	    if { $config(sendtoorigin) == 1 } {
		puts $c "$data"
	    } else {
		if { $chan != $c } {
		    puts $c "$data"
		}
	    }
	}
    } else {
	if { $config(debug) > 0 } {
	    puts stderr "Deleting client $chan from host $client and port $clientport."
	}
	DelClient $chan $client $clientport
    }
}

# NewClient --

# Sets up newly created connection after running access filters

# Arguments:

# chan: open channel

# client: client address

# clientport: port number of client

proc multiplexer::NewClient { chan client clientport } {
    variable clients
    variable config
    variable accessfilters
    # run through access filters
    foreach af $accessfilters {
	if { [$af $chan $client $clientport] == -1 } {
	    close $chan
	    return
	}
    }
    set clients($chan) $client
#    fconfigure $chan -blocking 0
    fconfigure $chan -buffering $config(buffering)
    fileevent $chan readable [list ::multiplexer::GetData $chan $client $clientport ]
    if { $config(debug) > 0 } {
	puts stderr "Tcl channel $chan is host $client and port $clientport."
    }
}

# Config --

# Configure global options, which currently include the following:

# sendtoorigin: if 1, resend the data to all clients, including the
# sender.  Defaults to 0

# buffering: fed to Tcl's fconfigure -buffering.  Defaults to line.

# debug: if greater than 0, prints additional debug messages.
# Defaults to 0

# Arguments:

# key: name of option to configure

# value: value for option.

proc multiplexer::Config { key value } {
    variable config
    set config($key) $value
}

# Init --

# Start the server

# Arguments:

# port: port to listen on.

proc multiplexer::Init { port } {
    socket -server ::multiplexer::NewClient $port
}

namespace eval multiplexer {
    namespace export Init Config AddAccessFilter AddFilter AddExitFilter
}